/*------------------------------------------------------------------------------
* rtksvr.c : rtk positioning server functions
*
*          Copyright (C) 2007-2009 by T.TAKASU, All rights reserved.
*
* version : $Revision:$ $Date:$
* history : 2009/01/07  1.0 new
*-----------------------------------------------------------------------------*/
#include "rtklib.h"

static const char rcsid[]="$Id:$";

/* write solution header to output stream ------------------------------------*/
static void writesolhead(int format, stream_t *stream)
{
    unsigned char buff[1024];
    int n;
    
    setsolformat(format);
    n=outsolheads(buff);
    strwrite(stream,buff,n);
}
/* save output buffer --------------------------------------------------------*/
static void saveoutbuf(rtksvr_t *svr, unsigned char *buff, int n, int index)
{
    rtksvrlock(svr);
    
    n=n<svr->buffsize-svr->nsb[index]?n:svr->buffsize-svr->nsb[index];
    memcpy(svr->sbuf[index]+svr->nsb[index],buff,n);
    svr->nsb[index]+=n;
    
    rtksvrunlock(svr);
}
/* write solution to output stream -------------------------------------------*/
static void writesol(rtksvr_t *svr)
{
    rtksol_t *sol;
    double rr[6],Qr[9],az[MAXOBS],el[MAXOBS];
    unsigned char buff[1024];
    int i,j,n,ns,sat[MAXOBS],snr[MAXOBS],vsat[MAXOBS];
    
    tracet(4,"writesol:\n");
    
    for (i=0;i<3;i++) {
        rr[i]=svr->rtkstat==1?svr->rtk.xa[i]:svr->rtk.x[i];
        rr[i+3]=svr->rtk.rr[0][i+3];
        for (j=0;j<3;j++) {
            Qr[i+j*3]=svr->rtkstat==1?
                svr->rtk.Pa[i+j*svr->rtk.na]:svr->rtk.P[i+j*svr->rtk.nx];
        }
    }
    for (i=0;i<2;i++) {
        /* 1Hz for nmea */
        if (svr->format[i+2]==SOLF_NMEA&&svr->rtkstat!=SOLQ_NONE&&
            DTTOL<svr->rtk.time[0].sec&&svr->rtk.time[0].sec<=1.0-DTTOL) continue;
        
        /* output solution */
        setsolformat(svr->format[i+2]);
        n=outsols(buff,svr->rtk.time[0],rr,Qr,svr->rtk.rr[1],svr->rtkstat,
                  svr->rtk.ns,svr->rtk.age,svr->rtk.ambratio);
        strwrite(svr->stream+i+2,buff,n);
        
        /* save output buffer */
        saveoutbuf(svr,buff,n,i);
        
        for (j=0;j<svr->obs[0].n;j++) {
            sat [j]=svr->obs[0].data[j].sat;
            snr [j]=svr->obs[0].data[j].SNR[0];
            az  [j]=svr->rtk.az[0][sat[j]-1];
            el  [j]=svr->rtk.el[0][sat[j]-1];
            vsat[j]=svr->rtk.vs[0][sat[j]-1];
        }
        /* output extended solution */
        n=outsolexs(buff,svr->rtkstat,svr->obs[0].n,sat,az,el,snr,vsat);
        strwrite(svr->stream+i+2,buff,n);
        
        /* save output buffer */
        saveoutbuf(svr,buff,n,i);
    }
    /* save solution buffer */
    if (svr->nsol<MAXSOLBUF) {
        rtksvrlock(svr);
        sol=svr->solbuf+svr->nsol;
        sol->stat=svr->rtkstat;
        sol->time=svr->rtk.time[0];
        for (i=0;i<6;i++) sol->rr[i]=rr[i];
        for (i=0;i<9;i++) sol->Qr[i]=Qr[i];
        for (i=0;i<3;i++) sol->rb[i]=svr->rtk.rr[1][i];
        sol->ns=svr->rtk.ns;
        sol->age=svr->rtk.age;
        sol->ratio=svr->rtk.ambratio;
        svr->nsol++;
        rtksvrunlock(svr);
    }
}
/* update rtk server struct --------------------------------------------------*/
static void updatesvr(rtksvr_t *svr, int ret, obs_t *obs, nav_t *nav, int sat,
                      sbsmsg_t *sbsmsg, int index)
{
    double rr[3],pos[3],enu[3]={0},dr[3];
    int i,n=0;
    
    tracet(4,"updatesvr: ret=%d index=%d\n",ret,index);
    
    if (ret==1) { /* observation data */
        for (i=0;i<obs->n;i++) {
            if (svr->rtk.opt.exsats[obs->data[i].sat-1]) continue;
            svr->obs[index].data[n]=obs->data[i];
            svr->obs[index].data[n++].rcv=index+1;
        }
        svr->obs[index].n=n;
        sortobs(svr->obs+index);
        svr->nmsg[index][0]++;
    }
    else if (ret==2) { /* ephemeris */
        if (svr->navsel==index||svr->navsel==2) {
            svr->nav.eph[sat-1]=nav->eph[sat-1];
        }
        svr->nmsg[index][1]++;
    }
    else if (ret==3) { /* sbas message */
        if (sbsmsg&&(svr->sbssat==sbsmsg->prn||svr->sbssat==0)) {
            if (svr->nsbs<MAXSBSMSG) {
                svr->sbsmsg[svr->nsbs++]=*sbsmsg;
            }
            else {
                for (i=0;i<MAXSBSMSG-1;i++) svr->sbsmsg[i]=svr->sbsmsg[i+1];
                svr->sbsmsg[i]=*sbsmsg;
            }
            sbsupdatestat(sbsmsg,&svr->sbsstat);
        }
        svr->nmsg[index][3]++;
    }
    else if (ret==9) { /* ion/utc parameters */
        if (svr->navsel==index||svr->navsel==2) {
            for (i=0;i<8;i++) svr->nav.ion[i]=nav->ion[i];
            for (i=0;i<4;i++) svr->nav.utc[i]=nav->utc[i];
            svr->nav.leaps=nav->leaps;
        }
        svr->nmsg[index][2]++;
    }
    else if (ret==5) { /* station postion parameters */
        if (svr->rtk.opt.refpos==3&&index==1) {
            for (i=0;i<3;i++) {
                rr[i]=svr->rtcm[1].sta.rr[i]+svr->rtcm[1].sta.del[0][i];
            }
            /* antenna height */
            ecef2pos(rr,pos);
            enu[2]=svr->rtcm[1].sta.anth;
            enu2ecef(pos,enu,dr);
            
            /* set base station position */
            for (i=0;i<3;i++) {
                svr->rtk.rr[1][i]=rr[i]+dr[i];
            }
        }
        svr->nmsg[index][4]++;
    }
    else if (ret==7) { /* dgps correction */
        svr->nmsg[index][5]++;
    }
    else if (ret==-1) { /* error */
        svr->nmsg[index][9]++;
    }
}
/* decode receiver raw/rtcm data ---------------------------------------------*/
static int decoderaw(rtksvr_t *svr, int index)
{
    obs_t *obs;
    nav_t *nav;
    sbsmsg_t *sbsmsg=NULL;
    int i,ret,sat,fobs=0;
    
    tracet(4,"decoderaw: index=%d\n",index);
    
    rtksvrlock(svr);
    
    for (i=0;i<svr->nb[index];i++) {
        
        /* input rtcm/receiver raw data from stream */
        if (svr->format[index]==STRFMT_RTCM2) {
            ret=input_rtcm2(svr->rtcm+index,svr->buff[index][i]);
            obs=&svr->rtcm[index].obs;
            nav=&svr->rtcm[index].nav;
            sat=svr->rtcm[index].ephsat;
        }
        else if (svr->format[index]==STRFMT_RTCM3) {
            ret=input_rtcm3(svr->rtcm+index,svr->buff[index][i]);
            obs=&svr->rtcm[index].obs;
            nav=&svr->rtcm[index].nav;
            sat=svr->rtcm[index].ephsat;
        }
        else {
            ret=input_raw(svr->raw+index,svr->format[index],svr->buff[index][i]);
            obs=&svr->raw[index].obs;
            nav=&svr->raw[index].nav;
            sat=svr->raw[index].ephsat;
            sbsmsg=&svr->raw[index].sbsmsg;
        }
        /* update rtk server */
        if (ret>0) updatesvr(svr,ret,obs,nav,sat,sbsmsg,index);
        
        /* observation data received */
        if (ret==1) fobs++;
    }
    svr->nb[index]=0;
    
    rtksvrunlock(svr);
    
    return fobs;
}
/* send nmea request ---------------------------------------------------------*/
static void sendnmea(stream_t *stream, double *pos)
{
    char buff[1024];
    int n;
    
    tracet(3,"sendnmea: nmeapos=%.3f %.3f %.3f\n",pos[0],pos[1],pos[2]);
    
    n=outnmea_gga(buff,utc2gpst(timeget()),pos,SOLQ_SINGLE,10,0.0);
    strwrite(stream,buff,n);
}
/* rtk server thread ---------------------------------------------------------*/
#ifdef WIN32
static DWORD WINAPI rtksvrthread(void *arg)
#else
static void *rtksvrthread(void *arg)
#endif
{
    rtksvr_t *svr=(rtksvr_t *)arg;
    obs_t obs;
    obsd_t data[MAXOBS*2];
    double pos[3];
    unsigned int tick,ticknmea;
    unsigned char *p,*q;
    int i,j,n,nobs,fobs[2],cycle,cputime;
    
    tracet(3,"rtksvrthread:\n");
    
    svr->state=1; obs.data=data;
    svr->tick=tickget();
    ticknmea=svr->tick-1000;
    
    for (cycle=0;svr->state;cycle++) {
        tick=tickget();
        
        for (i=0;i<2;i++) {
            p=svr->buff[i]+svr->nb[i]; q=svr->buff[i]+svr->buffsize;
            
            /* read receiver raw/rtcm data from input stream */
            if ((n=strread(svr->stream+i,p,q-p))<=0) {
                continue;
            }
            /* write receiver raw/rtcm data to log stream */
            strwrite(svr->stream+i+4,p,n);
            svr->nb[i]+=n;
            
            /* save peek buffer */
            rtksvrlock(svr);
            n=n<svr->buffsize-svr->npb[i]?n:svr->buffsize-svr->npb[i];
            memcpy(svr->pbuf[i]+svr->npb[i],p,n);
            svr->npb[i]+=n;
            rtksvrunlock(svr);
        }
        for (i=0;i<2;i++) { /* decode receiver raw/rtcm data */
            fobs[i]=decoderaw(svr,i);
        }
        if (fobs[0]) { /* rover observation data received */
            obs.n=0;
            for (i=0;i<svr->obs[0].n;i++) obs.data[obs.n++]=svr->obs[0].data[i];
            for (i=0;i<svr->obs[1].n;i++) obs.data[obs.n++]=svr->obs[1].data[i];
            
            /* rtk positioning */
            rtksvrlock(svr);
            svr->rtkstat=rtkpos(&svr->rtk,obs.data,obs.n,&svr->nav,&svr->sbsstat);
            rtksvrunlock(svr);
            
            /* write solution */
            if (svr->rtkstat!=SOLQ_NONE) writesol(svr);
            
            /* if multiple obs data, increment missing obs counter */
            if (fobs[0]>1) svr->prcout+=fobs[0]-1;
        }
        /* send null solution if no solution */
        if (svr->rtkstat==SOLQ_NONE&&cycle%(1000/svr->cycle)==0) {
            writesol(svr);
        }
        /* send nmea request to base/nrtk input stream */
        if (svr->nmeacycle>0&&(int)(tick-ticknmea)>=svr->nmeacycle) {
            if (svr->stream[1].state==1) {
                if (svr->nmeareq==1) {
                    sendnmea(svr->stream+1,svr->nmeapos);
                }
                else if (svr->nmeareq==2&&norm(svr->rtk.rr[0],3)>0.0) {
                    sendnmea(svr->stream+1,svr->rtk.rr[0]);
                }
            }
            ticknmea=tick;
        }
        if ((cputime=(int)(tickget()-tick))>0) svr->cputime=cputime;
        
        /* sleep until next cycle */
        sleepms(svr->cycle-cputime);
    }
    for (i=0;i<MAXSTRRTK;i++) strclose(svr->stream+i);
    for (i=0;i<2;i++) {
        svr->nb[i]=svr->nsb[i]=svr->npb[i]=0;
        free(svr->buff[i]); svr->buff[i]=NULL;
        free(svr->sbuf[i]); svr->sbuf[i]=NULL;
        free(svr->pbuf[i]); svr->pbuf[i]=NULL;
    }
    return 0;
}
/* initialize rtk server -------------------------------------------------------
* initialize rtk server
* args   : rtksvr_t *svr    IO rtk server
* return : status (0:error,1:ok)
*-----------------------------------------------------------------------------*/
extern int rtksvrinit(rtksvr_t *svr)
{
    rtksvr_t svr0={0};
    eph_t eph0={0};
    int i,j; 
    
    tracet(3,"rtksvrinit:\n");
    
    *svr=svr0;
    
    if (!(svr->nav.eph=(eph_t *)malloc(sizeof(eph_t)*MAXSAT))) return 0;
    for (j=0;j<MAXSAT;j++) svr->nav.eph[j]=eph0;
    svr->nav.n=MAXSAT;
    
    for (i=0;i<2;i++) {
        if (!(svr->obs[i].data=(obsd_t *)malloc(sizeof(obsd_t)*MAXOBS))||
            !init_raw(svr->raw+i)||!init_rtcm(svr->rtcm+i)) {
            for (i=0;i<2;i++) {
                free(svr->obs[i].data);
                free_raw(svr->raw+i);
                free_rtcm(svr->rtcm+i);
            }
            free(svr->nav.eph);
            return 0;
        }
    }
    for (i=0;i<MAXSTRRTK;i++) strinit(svr->stream+i);
    
    initlock(&svr->lock);
    
    return 1;
}
/* lock/unlock rtk server ------------------------------------------------------
* lock/unlock rtk server
* args   : rtksvr_t *svr    IO rtk server
* return : status (1:ok 0:error)
*-----------------------------------------------------------------------------*/
extern void rtksvrlock  (rtksvr_t *svr) {lock  (&svr->lock);}
extern void rtksvrunlock(rtksvr_t *svr) {unlock(&svr->lock);}

/* start rtk server ------------------------------------------------------------
* start rtk server thread
* args   : rtksvr_t *svr    IO rtk server
*          int     cycle    I  server cycle (ms)
*          int     buffsize I  input buffer size (bytes)
*          int     *strs    I  input/output/log stream types (STR_???)
*                              types[0]=input stream rover
*                              types[1]=input stream base station
*                              types[2]=output stream solution 1
*                              types[3]=output stream solution 2
*                              types[4]=log stream rover
*                              types[5]=log stream base station
*          char    *paths   I  input/output/log stream paths
*          int     *format  I  input/output/log stream formats (STRFMT_???)
*          int     navsel   I  navigation message select (0:rover,1:base,2:both)
*          int     sbssat   I  sbas satellite select (0:all,other:prn)
*          char    **cmds   I  input stream start commands
*                              cmds[0]=input stream rover
*                              cmds[1]=input stream base station
*          int     nmeacycle I nmea request cycle (ms) (0:no request)
*          int     nmeareq  I  nmea request type (0:no,1:base pos,2:single sol)
*          double *nmeapos  I  transmitted nmea position (ecef) (m)
*          prcopt_t *prcopt I  rtk processing options
*          solopt_t *solopt I  solution options
* return : status (1:ok 0:error)
*-----------------------------------------------------------------------------*/
extern int rtksvrstart(rtksvr_t *svr, int cycle, int buffsize, int *strs,
                       char **paths, int *formats, int navsel, int sbssat,
                       char **cmds, int nmeacycle, int nmeareq,
                       const double *nmeapos, prcopt_t *prcopt,
                       solopt_t *solopt)
{
    char cmdend[]="\r\n";
    int i,j,rw;
    
    tracet(3,"rtksvrstart:\n");
    
    if (svr->state) return 0;
    
    strinitcom();
    svr->cycle=cycle>1?cycle:1;
    svr->nmeacycle=nmeacycle>1000?nmeacycle:1000;
    svr->nmeareq=nmeareq;
    for (i=0;i<3;i++) svr->nmeapos[i]=nmeapos[i];
    svr->buffsize=buffsize>4096?buffsize:4096;
    for (i=0;i<4;i++) svr->format[i]=formats[i];
    svr->navsel=navsel;
    svr->sbssat=sbssat;
    svr->nsbs=0;
    svr->nsol=0;
    svr->rtkstat=0;
    svr->prcout=0;
    rtkfree(&svr->rtk);
    rtkinit(&svr->rtk,prcopt);
    
    for (i=0;i<2;i++) {
        svr->nb[i]=0;
        if (!(svr->buff[i]=(unsigned char *)malloc(buffsize))||
            !(svr->sbuf[i]=(unsigned char *)malloc(buffsize))||
            !(svr->pbuf[i]=(unsigned char *)malloc(buffsize))) {
            for (i--;i>=0;i--) {
                free(svr->buff[i]); svr->buff[i]=NULL;
                free(svr->sbuf[i]); svr->sbuf[i]=NULL;
                free(svr->pbuf[i]); svr->pbuf[i]=NULL;
            }
            return 0;
        }
        for (j=0;j<10;j++) svr->nmsg[i][j]=0;
        svr->obs[i].n=0;
        
        /* initialize receiver raw and rtcm control */
        init_raw (svr->raw +i);
        init_rtcm(svr->rtcm+i);
    }
    /* set solution options */
    setsolopt(solopt);
    
    /* set base station position */
    for (i=0;i<3;i++) {
        svr->rtk.rr[1][i]=prcopt->rb[i];
    }
    /* open streams */
    for (i=0;i<MAXSTRRTK;i++) {
        rw=i<2?STR_MODE_R:STR_MODE_W;
        if (strs[i]!=STR_FILE) rw|=STR_MODE_W;
        if (!stropen(svr->stream+i,strs[i],rw,paths[i])) {
            for (i--;i>=0;i--) strclose(svr->stream+i);
            return 0;
        }
        if (i<2&&strs[i]==STR_FILE) {
            svr->raw [i].time=strgettime(svr->stream+i);
            svr->rtcm[i].time=strgettime(svr->stream+i);
        }
    }
    /* sync input streams */
    strsync(svr->stream,svr->stream+1);
    
    /* write start commands to input streams */
    for (i=0;i<2;i++) {
        if (!cmds[i]) continue;
        strwrite(svr->stream+i,(unsigned char *)cmds[i],strlen(cmds[i]));
        strwrite(svr->stream+i,(unsigned char *)cmdend,2);
    }
    /* write solution header to solution streams */
    for (i=2;i<4;i++) {
        writesolhead(svr->format[i],svr->stream+i);
    }
    /* create rtk server thread */
#ifdef WIN32
    if (!(svr->thread=CreateThread(NULL,0,rtksvrthread,svr,0,NULL))) {
#else
    if (pthread_create(&svr->thread,rtksvrthread,svr))) {
#endif
        for (i=0;i<MAXSTRRTK;i++) strclose(svr->stream+i);
        return 0;
    }
    return 1;
}
/* stop rtk server -------------------------------------------------------------
* start rtk server thread
* args   : rtksvr_t *svr    IO rtk server
*          char    **cmds   I  input stream stop commands
*                              cmds[0]=input stream rover
*                              cmds[1]=input stream base station
* return : none
*-----------------------------------------------------------------------------*/
extern void rtksvrstop(rtksvr_t *svr, char **cmds)
{
    char cmdend[]="\r\n";
    int i;
    
    tracet(3,"rtksvrstop:\n");
    
    /* write stop commands to input streams */
    rtksvrlock(svr);
    for (i=0;i<2;i++) {
        if (!cmds[i]) continue;
        strwrite(svr->stream+i,(unsigned char *)cmds[i],strlen(cmds[i]));
        strwrite(svr->stream+i,(unsigned char *)cmdend,2);
    }
    rtksvrunlock(svr);
    
    /* stop rtk server */
    svr->state=0;
    
    /* free rtk server thread */
#ifdef WIN32
    WaitForSingleObject(svr->thread,10000);
    CloseHandle(svr->thread);
#else
    pthread_join(&svr->thread,NULL);
#endif
}
/* get observation data status -------------------------------------------------
* get current observation data status
* args   : rtksvr_t *svr    I  rtk server
*          int     rcv      I  receiver (0:rover,1:base station)
*          gtime_t *time    O  time of observation data
*          int     *sat     O  satellite prn numbers
*          double  *az      O  satellite azimuth angles (rad)
*          double  *el      O  satellite elevation angles (rad)
*          int     *snr1    O  satellite L1 snr (dBHz)
*          int     *snr2    O  satellite L2 snr (dBHz)
*          int     *vsat    O  valid satellite flag
* return : number of satellites
*-----------------------------------------------------------------------------*/
extern int rtksvrostat(rtksvr_t *svr, int rcv, gtime_t *time, int *sat,
                       double *az, double *el, int *snr1, int *snr2, int *vsat)
{
    int i,ns;
    
    tracet(4,"rtksvrostat:\n");
    
    if (!svr->state) return 0;
    rtksvrlock(svr);
    ns=svr->obs[rcv].n;
    if (ns>0) {
        *time=svr->obs[rcv].data[0].time;
    }
    for (i=0;i<ns;i++) {
        sat[i]=svr->obs[rcv].data[i].sat;
        az[i]=svr->rtk.az[0][sat[i]-1];
        el[i]=svr->rtk.el[0][sat[i]-1];
        snr1[i]=svr->obs[rcv].data[i].SNR[0];
        snr2[i]=svr->obs[rcv].data[i].SNR[1];
        if (svr->rtk.opt.mode>PMODE_SINGLE) {
            vsat[i]=svr->rtk.vsat[sat[i]-1][0];
        }
        else {
            vsat[i]=svr->rtk.vs[0][sat[i]-1];
        }
    }
    rtksvrunlock(svr);
    return ns;
}
/* get stream status -----------------------------------------------------------
* get current stream status
* args   : rtksvr_t *svr    I  rtk server
*          int     *sstat   O  status of streams
*          char    *msg     O  status messages
* return : none
*-----------------------------------------------------------------------------*/
extern void rtksvrsstat(rtksvr_t *svr, int *sstat, char *msg)
{
    int i;
    char s[MAXSTRMSG],*p=msg;
    
    tracet(4,"rtksvrsstat:\n");
    
    rtksvrlock(svr);
    for (i=0;i<MAXSTRRTK;i++) {
        sstat[i]=strstat(svr->stream+i,s);
        if (*s) p+=sprintf(p,"(%d) %s ",i+1,s);
    }
    rtksvrunlock(svr);
}
